'''
Send text to OBS Studio
'''

import logging
import simpleobsws
from slidetextbridge.core import config
from . import base

class ObsWsEmitter(base.PluginBase):
    '''
    Emit text to OBS Studio
    '''
    @staticmethod
    def type_name():
        return 'obsws'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('url', type=str, default='ws://localhost:4455/')
        cfg.add_argment('password', type=str, default=None)
        cfg.add_argment('source_name', type=str, default=None)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'obsws({self.cfg.location})')
        self.ws = None
        self.connect_to(cfg.src)

    async def _ws_connect(self):
        self.ws = simpleobsws.WebSocketClient(url=self.cfg.url, password=self.cfg.password)
        await self.ws.connect()
        if not await self.ws.wait_until_identified():
            self.ws.disconnect()
            self.ws = None
            raise ConnectionError('Identification error')

    async def update(self, slide, args):
        if not slide:
            text = ''
        elif isinstance(slide, str):
            text = slide
        else:
            text = str(slide)

        try:
            await self._send_request('SetInputSettings', {
                'inputName': self.cfg.source_name,
                'inputSettings': {'text': text}
            })
        except Exception as e:
            self.logger.warning('Could not send text. %s', e)
            return

    async def _send_request(self, req, data, retry=2):
        while retry > 0:
            retry -= 1

            if not self.ws:
                try:
                    await self._ws_connect()
                except Exception as e:
                    self.logger.warning('Could not connect to %s. %s', self.cfg.url, e)
                    self.ws = None
            if not self.ws:
                if retry > 0:
                    self.logger.info('Retrying to connect...')
                continue

            try:
                res = await self.ws.call(simpleobsws.Request(req, data))
                if res.ok():
                    return res.responseData
                self.logger.warning('%s: %d: %s', res.requestType,
                                    res.requestStatus.code, res.requestStatus.comment)

            except Exception as e:
                self.logger.warning('Failed to send text. %s', e)

                try:
                    await self.ws.disconnect()
                except Exception as e1:
                    self.logger.warning('Disconnect failed. %s', e1)
                self.ws = None

                if retry == 0:
                    raise e
                self.logger.info('Retrying...')



'''
Get text from Microsoft PowerPoint
'''

import asyncio
import logging
import win32com.client
import pywintypes
from slidetextbridge.core import config
from slidetextbridge.core.logging import HideSameLogFilter
from . import base


class _Const:
    # pylint: disable=R0903
    ppSlideShowBlackScreen = 3
    ppSlideShowWhiteScreen = 4
    ppSlideShowDone = 5


class PowerPointCapture(base.PluginBase):
    '''
    Get text from Microsoft PowerPoint
    '''
    @staticmethod
    def type_name():
        return 'ppt'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg, has_src=False)
        cfg.add_argment('placeholder_only', type=bool, default=True)
        cfg.add_argment('poll_wait_time', type=float, default=0.1)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'ppt({self.cfg.location})')
        self.logger.addFilter(HideSameLogFilter())
        self._last_slide = self
        self.ppt = None
        self._last_window = None

    async def initialize(self):
        asyncio.create_task(self._loop())

    async def _loop(self):
        while True:
            try:
                await self._loop_once()
            except Exception as e:
                self.logger.error('Unknown error: %s', e)
                await asyncio.sleep(1)

    async def _loop_once(self):
        slide = self._get_slide()
        if slide != self._last_slide:
            self._last_slide = slide
            await self.emit(PowerPointSlide(slide, cfg=self.cfg))

        await asyncio.sleep(self.cfg.poll_wait_time)

    def _connect_ppt(self):
        self.logger.info('Connecting to PowerPoint...')
        try:
            self.ppt = win32com.client.Dispatch('PowerPoint.Application')
        except pywintypes.com_error as e:
            self.logger.error('Failed to dispatch PowerPoint. Check PowerPoint installation. %s', e)
            self.ppt = None
            return
        self._last_window = None
        if self.ppt:
            self.logger.info('Connected to PowerPoint.')
        else:
            self.logger.warning('Failed to connect to PowerPoint.')

    def _current_slideshow_window(self):
        if not self.ppt:
            self._connect_ppt()
            if not self.ppt:
                return None
        try:
            slide_count = self.ppt.SlideShowWindows.Count
        except (pywintypes.com_error, AttributeError): # pylint: disable=no-member
            self.ppt = None
            self._last_window = None
            return None

        if slide_count == 0:
            self.logger.warning('No current presentation window.')
            return None

        if slide_count == 1:
            self._last_window = self.ppt.SlideShowWindows(1)
            return self._last_window

        cand = None
        for ix in range(slide_count):
            w = self.ppt.SlideShowWindows(ix + 1)
            if w.Active:
                self._last_window = w
                return w
            if w == self._last_window:
                cand = w
        return cand

    def _get_slide(self):
        w = self._current_slideshow_window()
        if not w:
            # no current presentation window
            return None

        view = w.View

        if view.State in (
                _Const.ppSlideShowBlackScreen,
                _Const.ppSlideShowWhiteScreen,
                _Const.ppSlideShowDone):
            return None

        return view.Slide


def _com32bool_to_bool(b):
    return bool(b)

def _font_to_dict(font):
    return base.SlideBase.convert_object(font, params=(
        'Size', 'Bold', 'Name', 'BaselineOffset',
        ('Italic', _com32bool_to_bool),
        ('Subscript', _com32bool_to_bool),
        ('Superscript', _com32bool_to_bool),
    ))

def _tr_to_dict(tr):
    return base.SlideBase.convert_object(
            tr,
            params=(
                ('HasText', _com32bool_to_bool),
                'Text',
                'Count', 'Start', 'Length',
                'BoundLeft', 'BoundTop', 'BoundWidth', 'BoundHeight',
                ('Font', _font_to_dict),
            ),
    )

def _tf_to_dict(tf):
    return base.SlideBase.convert_object(tf, params=(
        ('HasText', _com32bool_to_bool),
        ('TextRange', _tr_to_dict),
        'Orientation',
        ('WordWrap', _com32bool_to_bool),
    ))

def _pf_to_dict(pf):
    return base.SlideBase.convert_object(pf, params=('Name', 'Type', 'ContainedType'))

class PowerPointSlide(base.SlideBase):
    'The slide class returned by PowerPointCapture'

    def __init__(self, slide=None, data=None, cfg=None):
        self._slide = slide
        self.cfg = cfg
        if isinstance(data, list):
            self._dict = {'shapes': data}
        else:
            self._dict = data

    def _shape_is_valid(self, shape):
        if not shape.HasTextFrame or not shape.TextFrame.HasText:
            return False
        if self.cfg and self.cfg.placeholder_only:
            if shape.Type != 14:
                return False
        return True

    def to_texts(self):
        '''
        List all texts
        :return:  List of strings
        '''
        if self._dict:
            return _list_texts(self._dict)

        if not self._slide:
            return []

        texts = []
        for shape in self._slide.Shapes:
            if not self._shape_is_valid(shape):
                continue
            texts.append(shape.TextFrame.TextRange.Text.replace('\r', ' '))
        return texts

    def to_dict(self):
        if self._dict:
            return self._dict
        if not self._slide:
            return {}
        shapes = []
        for shape in self._slide.Shapes:
            if not self._shape_is_valid(shape):
                continue
            s = base.SlideBase.convert_object(shape, params=(
                ('TextFrame', _tf_to_dict),
                'Type',
                ('PlaceholderFormat', _pf_to_dict),
                'Name',
            ))
            shapes.append(s)
        self._dict = {'shapes': shapes}
        return self._dict


def _list_texts(obj):
    if isinstance(obj, str):
        return [obj, ]
    if isinstance(obj, (int, float, bool)):
        return []
    if isinstance(obj, dict):
        for key in ('shapes', 'text', 'text_range', 'text_frame'):
            if key in obj:
                return _list_texts(obj[key])
        return []
    ret = []
    for x in obj:
        if x:
            ret += _list_texts(x)
    return ret



'''
Accumulate plugins
'''

import os
import logging
from . import base
from . import text_filters
from . import debug_helper

logger = logging.getLogger(__name__)

_is_win = os.name.startswith('nt')

plugin_classes: list[type[base.PluginBase]] = []

plugin_classes.append(text_filters.TextLinebreakFilter)
plugin_classes.append(text_filters.RegexFilter)
plugin_classes.append(debug_helper.StdoutEmitter)

try:
    from . import obsws
    plugin_classes.append(obsws.ObsWsEmitter)
except ImportError as e:
    logger.info('obsws is unsupported. %s', e)

if _is_win:
    try:
        from . import powerpoint
        plugin_classes.append(powerpoint.PowerPointCapture)
    except ImportError as e:
        logger.info('Microsoft PowerPoint is unsupported %s', e)

try:
    from . import impress
    plugin_classes.append(impress.ImpressCapture)
except ImportError as e:
    logger.info('LibreOffice is unsupported %s', e)

try:
    from . import openlp
    plugin_classes.append(openlp.OpenLPCapture)
except ImportError as e:
    logger.info('OpenLP is unsupported %s', e)

try:
    from . import jmespath_filter
    plugin_classes.append(jmespath_filter.JMESPathFilter)
except ImportError as e:
    logger.info('JMESPath filter is unsupported %s', e)

try:
    from . import webserver
    plugin_classes.append(webserver.WebServerEmitter)
except ImportError as e:
    logger.info('Web server is unsupported %s', e)

plugins = {cls.type_name(): cls for cls in plugin_classes}



'''
Get text from LibreOffice Impress
'''

import asyncio
import logging
import uno
from slidetextbridge.core import config
from slidetextbridge.core.logging import HideSameLogFilter
from . import base


class ImpressCapture(base.PluginBase):
    '''
    Get text from LibreOffice Impress
    '''
    @staticmethod
    def type_name():
        return 'impress'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg, has_src=False)
        cfg.add_argment('host', type=str, default='localhost')
        cfg.add_argment('port', type=int, default=2002)
        cfg.add_argment('pipe_name', type=str)
        cfg.add_argment('poll_wait_time', type=float, default=0.1)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'impress({self.cfg.location})')
        self.logger.addFilter(HideSameLogFilter(4))
        self._last_slide = self
        self._desktop = None

    async def initialize(self):
        asyncio.create_task(self._loop())

    async def _loop(self):
        while True:
            if not await self._loop_once():
                await asyncio.sleep(1)

    async def _loop_once(self):
        if not self._desktop:
            try:
                self._connect()
                self.logger.info('Connected to impress.')
            except Exception as e:
                self._desktop = None
                self.logger.warning('Failed to connect impress. %s', e)
                return False

        try:
            slide = self._get_slide()
        except Exception as e:
            self.logger.warning('Failed to get slide. %s', e)
            self._desktop = None
            return False

        if slide != self._last_slide:
            self._last_slide = slide
            await self.emit(ImpressSlide(slide))

        await asyncio.sleep(self.cfg.poll_wait_time)
        return True

    def _connect(self):
        context = uno.getComponentContext()
        resolver = context.ServiceManager.createInstanceWithContext(
                'com.sun.star.bridge.UnoUrlResolver', context)
        if self.cfg.pipe_name:
            uno_conn = f'uno:pipe,name={self.cfg.pipe_name}'
        else:
            uno_conn = f'uno:socket,host={self.cfg.host},port={self.cfg.port}'
        uno_inst = resolver.resolve(f'{uno_conn};urp;StarOffice.ComponentContext')
        self._desktop = uno_inst.ServiceManager.createInstanceWithContext(
                'com.sun.star.frame.Desktop', uno_inst)

    def _get_slide(self):
        if not self._desktop:
            self.logger.warning('No desktop is available')
            return None

        c = self._desktop.getCurrentComponent()
        if not c:
            self.logger.info('Found no component. Probably, all windows have been closed.')
            return None
        presentation = c.getPresentation()
        controller = presentation.getController()
        if not controller:
            self.logger.info('Found no active presentation controller.')
            return None
        return controller.getCurrentSlide()


class ImpressSlide(base.SlideBase):
    'The slide class returned by ImpressCapture'

    def __init__(self, slide=None, data=None):
        self._slide = slide
        if isinstance(data, list):
            self._dict = {'shapes': data}
        else:
            self._dict = data

    def to_texts(self):
        '''
        List all texts
        :return:  List of strings
        '''
        if self._dict:
            return _list_texts(self._dict)
        if self._slide:
            texts = []
            for shape in self._slide:
                texts.append(shape.Text.getString())
            return texts
        return []

    def to_dict(self):
        if self._dict:
            return self._dict
        if not self._slide:
            return {}
        shapes = []
        for shape in self._slide:
            s = base.SlideBase.convert_object(shape, params=(
                ('Text', lambda v: v.getString()),
                'CharColor',
                'CharHeight',
                'CharFontName',
            ))
            s['text'] = shape.Text.getString()
            shapes.append(s)
        self._dict = {'shapes': shapes}
        return self._dict


def _list_texts(obj):
    if isinstance(obj, str):
        return [obj, ]
    if isinstance(obj, (int, float, bool)):
        return []
    if isinstance(obj, dict):
        for key in ('shapes', 'text'):
            if key in obj:
                return _list_texts(obj[key])
        return []
    ret = []
    for x in obj:
        if x:
            ret += _list_texts(x)
    return ret



import sys
import types
import unittest
from unittest.mock import MagicMock, patch, AsyncMock

try:
    import uno
except ImportError:
    mock_uno = types.SimpleNamespace(getComponentContext=lambda: None)
    sys.modules['uno'] = mock_uno

from slidetextbridge.plugins import impress


class TestImpressCapture(unittest.IsolatedAsyncioTestCase):
    def test_type_name(self):
        self.assertEqual(impress.ImpressCapture.type_name(), 'impress')

    @patch('uno.getComponentContext')
    async def test_loop_once(self, MockGetComponentContext):
        ctx = MagicMock()

        uno_ctx = MagicMock()
        MockGetComponentContext.return_value = uno_ctx

        resolver = MagicMock()
        uno_ctx.ServiceManager.createInstanceWithContext.return_value = resolver

        uno_inst = MagicMock()
        resolver.resolve.return_value = uno_inst

        desktop = MagicMock()
        uno_inst.ServiceManager.createInstanceWithContext.return_value = desktop

        component = MagicMock()
        desktop.getCurrentComponent.return_value = component

        presentation = MagicMock()
        component.getPresentation.return_value = presentation

        controller = MagicMock()
        presentation.getController.return_value = controller

        uno_slide = MagicMock()
        controller.getCurrentSlide.return_value = uno_slide

        cfg = impress.ImpressCapture.config({})
        obj = impress.ImpressCapture(ctx=ctx, cfg=cfg)

        with patch('asyncio.sleep') as mock_sleep:
            mock_sleep.return_value = None

            obj.emit = AsyncMock()
            shape1 = MagicMock()
            shape1.Text.getString.return_value = 'text1'
            shape2 = MagicMock()
            shape2.Text.getString.return_value = 'text2'
            uno_slide.__iter__.return_value = [shape1, shape2]

            await obj._loop_once()

            slide = obj.emit.call_args[0][0]
            self.assertEqual(slide._slide, uno_slide)
            self.assertEqual(slide.to_texts(), ['text1', 'text2'])

            mock_sleep.assert_called_once()

        uno_ctx.ServiceManager.createInstanceWithContext.assert_called_once_with(
                'com.sun.star.bridge.UnoUrlResolver', uno_ctx)
        resolver.resolve.assert_called_once_with(
                'uno:socket,host=localhost,port=2002;urp;StarOffice.ComponentContext')
        uno_inst.ServiceManager.createInstanceWithContext.assert_called_once_with(
                'com.sun.star.frame.Desktop', uno_inst)

    def test_shape_to_dict(self):
        shape1 = MagicMock()
        shape1.Text.getString.return_value = 'text1'
        shape1.CharHeight = 16
        shape2 = MagicMock()
        shape2.Text.getString.return_value = 'text2'
        shape2.CharHeight = 16

        slide = impress.ImpressSlide(slide=[shape1, shape2])
        d = slide.to_dict()
        self.assertEqual(len(d['shapes']), 2)
        self.assertEqual(d['shapes'][0]['text'], 'text1')
        self.assertEqual(d['shapes'][0]['char_height'], 16)

        # cover to_texts()
        self.assertEqual(str(slide), 'text1\ntext2')

if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import MagicMock, AsyncMock

from slidetextbridge.plugins import text_filters

from test.filter_helper import *


class TestTextFilters(unittest.IsolatedAsyncioTestCase):

    def test_type_name(self):
        self.assertEqual(text_filters.TextLinebreakFilter.type_name(), 'linebreak')
        self.assertEqual(text_filters.RegexFilter.type_name(), 'regex')

    def test_config(self):
        cfg = text_filters.RegexFilter.config(
                data = {'patterns': [
                    {'p': 'p1', 'r': 'r1'},
                    {'p': 'p2', 'r': 'r2'},
                ]}
        )

        self.assertEqual(len(cfg.patterns), 2)
        self.assertEqual(cfg.patterns[0][0].pattern, 'p1')
        self.assertEqual(cfg.patterns[0][1], 'r1')
        self.assertEqual(cfg.patterns[1][0].pattern, 'p2')
        self.assertEqual(cfg.patterns[1][1], 'r2')

    async def test_linebreak_filter_strip(self):

        res = await run_filter(
                text_filters.TextLinebreakFilter,
                {'strip': True,},
                make_slide(['A', 'B ', ' C ', ' D', 'E\n\nF'])
        )
        self.assertEqual(res.to_texts(), ['A', 'B', 'C', 'D', 'E\nF'])

        res = await run_filter(
                text_filters.TextLinebreakFilter,
                {'strip': False,},
                make_slide(['A', 'B ', ' C ', ' D', 'E\n\nF'])
        )
        self.assertEqual(res.to_texts(), ['A', 'B ', ' C ', ' D', 'E\n\nF'])

    async def test_linebreak_filter_split_join(self):

        # Let's make a method that takes cfg, input_texts, and output_texts as the parameters.
        res = await run_filter(
                text_filters.TextLinebreakFilter,
                {
                    'split_long_line': 8,
                    'joined_column_max': 4,
                    'join_by': '-'
                },
                make_slide([
                    'abcdefghijk', # Check too long ASCII word
                    'a\nb',
                    '.'*12,
                ])
        )
        self.assertEqual(res.to_texts(), [
            'abcdefgh\nijk',
            'a-b',
            '.'*12,
        ])

        res = await run_filter(
                text_filters.TextLinebreakFilter,
                {
                    'split_long_line': 4,
                    'split_nowrap_allow_overflow': False,
                },
                make_slide([
                    'abc',
                    'abcd',
                    'abcde',
                    'abcdef',
                    '...',
                    '....',
                    '.....',
                    '......',
                ])
        )
        self.assertEqual(res.to_texts(), [
            'abc',
            'abcd',
            'abcd\ne',
            'abcd\nef',
            '...',
            '....',
            '.....',
            '......',
        ])

    async def test_linebreak_filter_delimiters(self):

        res = await run_filter(
                text_filters.TextLinebreakFilter,
                {
                    'shape_delimiter': '/',
                    'line_delimiter': ':',
                },
                make_slide([
                    'a\nb',
                    'c',
                ])
        )
        self.assertEqual(res.to_texts(), [
            'a:b',
            'c',
        ])
        self.assertEqual(str(res), 'a:b/c')

    async def test_regex_filter(self):

        res = await run_filter(
                text_filters.RegexFilter,
                {
                    'patterns': [
                        {'p': r'(Th|th)is', 'r': r'\1at'},
                    ],
                },
                slides=[
                    make_slide(['This is a pen.']),
                    make_slide(['Who is this?']),
                ],
        )
        self.assertEqual(res[0].to_texts(), ['That is a pen.'])
        self.assertEqual(res[1].to_texts(), ['Who is that?'])

    # TODO: Also check the combination with jmespath filter


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock, AsyncMock

from slidetextbridge.plugins import text_filters

from test.filter_helper import *

class TestTextFiltersJa(unittest.IsolatedAsyncioTestCase):

    async def test_linebreak_filter_split(self):
        ctx = MagicMock()

        # Check the punctuation will overflow.
        cfg = text_filters.TextLinebreakFilter.config({
            'split_long_line': 52,
            'split_nowrap': '、。」』,.',
            'split_nowrap_allow_overflow': True,
        })
        filter_obj = text_filters.TextLinebreakFilter(ctx=ctx, cfg=cfg)
        filter_obj.emit = AsyncMock()
        slide = make_slide([
            'そのとき、イエスは言われた、「父よ、彼らをおゆるしください。彼らは何をしているのか、わからずにいるのです」。'
            + '人々はイエスの着物をくじ引きで分け合った。\n'
            + '民衆は立って見ていた。',
            '。' * 28,
            ])

        await filter_obj.update(slide, args=None)

        filter_obj.emit.assert_awaited_once()
        res = filter_obj.emit.await_args[0][0].to_texts()
        self.assertEqual(res, ['\n'.join((
            'そのとき、イエスは言われた、「父よ、彼らをおゆるしく',
            'ださい。彼らは何をしているのか、わからずにいるのです」。',
            '人々はイエスの着物をくじ引きで分け合った。',
            '民衆は立って見ていた。'
            )), '。' * 28])

        # Check the punctuation won't overflow by wrapping at previous char.
        filter_obj.cfg.split_nowrap_allow_overflow = False
        filter_obj.emit = AsyncMock()

        await filter_obj.update(slide, args=None)

        filter_obj.emit.assert_awaited_once()
        res = filter_obj.emit.await_args[0][0].to_texts()
        self.assertEqual(res, ['\n'.join((
            'そのとき、イエスは言われた、「父よ、彼らをおゆるしく',
            'ださい。彼らは何をしているのか、わからずにいるので',
            'す」。人々はイエスの着物をくじ引きで分け合った。',
            '民衆は立って見ていた。'
            )), '。' * 28])

        # Check if there is a space.
        slide = make_slide([
            '34 そのとき、イエスは言われた、「父よ、彼らをおゆるしください。彼らは何をしているのか、わからずにいるのです」。'
            + '人々はイエスの着物をくじ引きで分け合った。'
            ])
        filter_obj.cfg.split_nowrap_allow_overflow = True
        filter_obj.emit = AsyncMock()

        await filter_obj.update(slide, args=None)

        filter_obj.emit.assert_awaited_once()
        res = filter_obj.emit.await_args[0][0].to_texts()
        self.assertEqual(res, ['\n'.join((
            '34 そのとき、イエスは言われた、「父よ、彼らをおゆる',
            'しください。彼らは何をしているのか、わからずにいるの',
            'です」。人々はイエスの着物をくじ引きで分け合った。',
            ))])

        # Check if there are non-CJK words and spaces in Japanese text.
        slide = make_slide([''.join((
            'Χριστοῦ Ἰησοῦ の僕、神の福音のために選び別たれ、召されて使徒となった Παῦλος から ',
            '―― この福音は、神が、預言者たちにより、聖書の中で、あらかじめ約束されたもので',
            'あって、 御子に関するものである。御子は、肉によれば Δαυὶδ の子孫から生れ、聖',
            'なる霊によれば、死人からの復活により、御力をもって神の御子と定められた。これがわ',
            'たしたちの主 Ἰησοῦ Χριστοῦ である。'
            )) ])
        filter_obj.emit = AsyncMock()
        filter_obj.cfg.split_long_line = 48

        await filter_obj.update(slide, args=None)

        filter_obj.emit.assert_awaited_once()
        res = filter_obj.emit.await_args[0][0].to_texts()
        self.assertEqual(res, ['\n'.join((
            'Χριστοῦ Ἰησοῦ の僕、神の福音のために選び別たれ、',
            '召されて使徒となった Παῦλος から ―― この福音は、',
            '神が、預言者たちにより、聖書の中で、あらかじめ約',
            '束されたものであって、 御子に関するものである。',
            '御子は、肉によれば Δαυὶδ の子孫から生れ、聖なる',
            '霊によれば、死人からの復活により、御力をもって神',
            'の御子と定められた。これがわたしたちの主 Ἰησοῦ', # <-- Hit the case the non-CJK word is at the end.
            'Χριστοῦ である。'
            ))])


if __name__ == '__main__':
    unittest.main()



'''
Base class definition
'''

import logging
import re

def set_config_arguments(cfg, has_src=True):
    '''
    Set the standard config arguments
    :param cfg:  Instance of ConfigBase
    '''
    cfg.add_argment('type', type=str)
    cfg.add_argment('name', type=str, default=None)
    if has_src:
        cfg.add_argment('src', type=str, default=None)


class PluginBase:
    '''
    The base class for text sources, filters, and sinks.
    '''

    @staticmethod
    def type_name():
        'Return the name of the type'
        return 'base'

    @staticmethod
    def config(data):
        'Return the config object'
        raise NotImplementedError()

    def __init__(self, ctx, cfg=None):
        self.ctx = ctx
        self.cfg = cfg
        self.name = cfg.name
        self.sinks = []
        if not hasattr(self, 'logger'):
            self.logger = logging.getLogger(f'{__name__}({self.cfg.location})')

    def connect_to(self, name=None, args=None):
        '''
        Connect this instance to the source of the text.
        :param name:  The name of the source
        :param args:  The argument that will be passed to `update` callback
        '''
        try:
            src = self.ctx.get_instance(name=name)
            return src.add_sink(self, args=args)
        except (IndexError, KeyError) as e:
            self.logger.error('Cannot connect to the source. %s', e)
            raise e

    def add_sink(self, sink, args=None):
        '''
        Connect a sink, that will receive the contents of this plugin.
        :param sink:  The instance of PluginBase
        :param args:  Any object
        '''
        self.sinks.append((sink, args))

    async def initialize(self):
        '''
        Callback when the system is started.
        '''

    async def emit(self, slide):
        '''
        Send the new slide to all sinks.
        :param slide:  The new slide.
        '''
        for sink, args in self.sinks:
            try:
                await sink.update(slide, args)
            except Exception as e:
                self.logger.error('Failed to send text to %s. %s', type(sink), e)

    async def update(self, slide, args):
        '''
        The callback function when updating the new slide.
        '''


class SlideBase():
    '''
    The base class representing the current slide
    '''
    def __init__(self, slide=None, data=None):
        pass

    def to_texts(self):
        '''
        List all texts
        :return:  List of strings
        '''
        return []

    def __str__(self):
        return '\n'.join(self.to_texts())

    @staticmethod
    def convert_object(obj, params=()):
        '''
        A helper method to convert an object to dict type.
        :param params:  List of attributes that will be converted to the dict item.
        Each element of `params` can be a string or a 2-element tuple.
        '''
        d = {}
        def _f(attrname, f_conv=None):
            dest = re.sub('(.)([A-Z])', r'\1_\2', attrname).lower()
            try:
                v = getattr(obj, attrname)
            except: # pylint: disable=W0702
                return
            if f_conv:
                v = f_conv(v)
            d[dest] = v
        for p in params:
            if isinstance(p, tuple):
                _f(p[0], p[1])
            else:
                _f(p)
        return d

    def to_dict(self):
        '''
        Get the dict object so that jmespath can filter it.
        The jmespath filter will then calls the initializer of the class
        with a `data` parameter to create a new slide object.
        :return: dict object representing this object
        '''
        return {}



import unittest
from unittest.mock import MagicMock, patch, AsyncMock
import asyncio

from slidetextbridge.plugins import obsws, base

class TestObsWsEmitter(unittest.IsolatedAsyncioTestCase):

    def test_type_name(self):
        self.assertEqual(obsws.ObsWsEmitter.type_name(), 'obsws')

    def test_config(self):
        cfg_data = {'url': 'ws://192.0.2.1:4455/', 'password': 'pw', 'source_name': 'TestSource'}
        cfg = obsws.ObsWsEmitter.config(cfg_data)
        self.assertEqual(cfg.url, 'ws://192.0.2.1:4455/')
        self.assertEqual(cfg.password, 'pw')
        self.assertEqual(cfg.source_name, 'TestSource')


    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_connect_and_update(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src = 'dummy'
        cfg_url = 'ws://localhost:4455/'
        cfg_password = 'secret'
        cfg_src_name = 'test-source-name'
        cfg = obsws.ObsWsEmitter.config({
                'src': cfg_src,
                'url': cfg_url,
                'password': cfg_password,
                'source_name': cfg_src_name
        })

        mock_ws = AsyncMock()
        mock_res = MagicMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.return_value = None
        mock_ws.wait_until_identified.return_value = True
        mock_ws.call.return_value = mock_res
        mock_res.ok.return_value = True
        mock_res.responseData = {'status': 'ok'}

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        ctx.get_instance.assert_called_once_with(name=cfg_src)

        await obj.update('test text', args=None)

        MockWebSocketClient.assert_called_with(url=cfg_url, password=cfg_password)
        mock_ws.connect.assert_awaited_once()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': 'test text'},
            }
        ))

        await obj.update('2nd text', args=None)

        mock_ws.connect.assert_awaited_once()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': '2nd text'},
            }
        ))

    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_update_failure(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src_name = 'test-source-name'
        cfg_url = 'ws://192.0.2.2:4455/'
        cfg = obsws.ObsWsEmitter.config({
                'url': cfg_url,
                'source_name': cfg_src_name,
        })

        mock_ws = AsyncMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.side_effect = (Exception('error test 1'), Exception('error test 2'))
        mock_ws.wait_until_identified.return_value = False

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        obj.logger.warning = MagicMock()
        ctx.get_instance.assert_called_once_with(name=None)

        await obj.update(None, args=None)

        MockWebSocketClient.assert_called_with(url=cfg_url, password=None)
        mock_ws.wait_until_identified.assert_not_called()
        mock_ws.connect.assert_awaited()
        self._assert_logging(obj.logger.warning, (
            ('Could not connect to %s. %s', cfg_url, 'error test 1'),
            ('Could not connect to %s. %s', cfg_url, 'error test 2'),
        ));

    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_identify_failure(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src_name = 'test-source-name'
        cfg_url = 'ws://192.0.2.2:4455/'
        cfg = obsws.ObsWsEmitter.config({
                'url': cfg_url,
                'source_name': cfg_src_name,
        })

        mock_ws = AsyncMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.wait_until_identified.side_effect = (False, True)
        mock_ws.call.return_value = MagicMock()
        mock_ws.call.return_value.ok.return_value = True
        mock_ws.call.return_value.responseData = {'status': 'ok'}

        mock_ws.disconnect = MagicMock()

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        obj.logger.warning = MagicMock()
        ctx.get_instance.assert_called_once_with(name=None)

        await obj.update('', args=None)

        MockWebSocketClient.assert_called_with(url=cfg_url, password=None)
        mock_ws.wait_until_identified.assert_awaited()
        mock_ws.connect.assert_awaited()
        mock_ws.disconnect.assert_called_with()
        self._assert_logging(obj.logger.warning, (
            ('Could not connect to %s. %s', cfg.url, 'Identification error'),
        ));

        # The 2nd attempt will successfully send the request.
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': ''},
            }
        ))


    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_update_retry_1(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src_name = 'test-source-name'
        cfg = obsws.ObsWsEmitter.config({
                'source_name': cfg_src_name
        })

        mock_ws = AsyncMock()
        mock_res = MagicMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.return_value = None
        mock_ws.wait_until_identified.return_value = True
        mock_ws.call.side_effect = (Exception('error 1'), mock_res)
        mock_res.ok.return_value = True
        mock_res.responseData = {'status': 'ok'}

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        obj.logger.warning = MagicMock()
        ctx.get_instance.assert_called_once_with(name=None)

        await obj.update('test text', args=None)

        MockWebSocketClient.assert_called_with(url='ws://localhost:4455/', password=None)
        mock_ws.connect.assert_awaited()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': 'test text'},
            }
        ))
        self._assert_logging(obj.logger.warning, (
            ('Failed to send text. %s', 'error 1'),
        ));

    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_update_retry_2(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src_name = 'test-source-name'
        cfg = obsws.ObsWsEmitter.config({
                'source_name': cfg_src_name
        })

        mock_ws = AsyncMock()
        mock_res = MagicMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.return_value = None
        mock_ws.wait_until_identified.return_value = True
        mock_ws.call.side_effect = (Exception('error 1'), Exception('error 2'), mock_res)
        mock_res.ok.return_value = True
        mock_res.responseData = {'status': 'ok'}

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        obj.logger.warning = MagicMock()
        ctx.get_instance.assert_called_once_with(name=None)

        await obj.update(base.SlideBase(), args=None)

        MockWebSocketClient.assert_called_with(url='ws://localhost:4455/', password=None)
        mock_ws.connect.assert_awaited()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': ''},
            }
        ))
        self._assert_logging(obj.logger.warning, (
            ('Failed to send text. %s', 'error 1'),
            ('Failed to send text. %s', 'error 2'),
            ('Could not send text. %s', 'error 2'),
        ))

    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_update_retry_connect_failure(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src_name = 'test-source-name'
        cfg = obsws.ObsWsEmitter.config({
                'source_name': cfg_src_name
        })

        mock_ws = AsyncMock()
        mock_res = MagicMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.side_effect = (None, Exception('connect error test'))
        mock_ws.wait_until_identified.return_value = True
        mock_ws.call.side_effect = (Exception('error 1'), mock_res)
        mock_res.ok.return_value = True
        mock_res.responseData = {'status': 'ok'}

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        obj.logger.warning = MagicMock()
        ctx.get_instance.assert_called_once_with(name=None)

        await obj.update(base.SlideBase(), args=None)

        MockWebSocketClient.assert_called_with(url='ws://localhost:4455/', password=None)
        mock_ws.connect.assert_awaited()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': ''},
            }
        ))
        self._assert_logging(obj.logger.warning, (
            ('Failed to send text. %s', 'error 1'),
            ('Could not connect to %s. %s', cfg.url, 'connect error test'),
        ))
        self.assertEqual(obj.ws, None)

    @patch('simpleobsws.WebSocketClient', autospec=True)
    async def test_wrong_source_name(self, MockWebSocketClient):
        ctx = MagicMock()

        cfg_src = 'dummy'
        cfg_url = 'ws://localhost:4455/'
        cfg_password = 'secret'
        cfg_src_name = 'test-source-name'
        cfg = obsws.ObsWsEmitter.config({
                'src': cfg_src,
                'url': cfg_url,
                'password': cfg_password,
                'source_name': cfg_src_name
        })

        mock_ws = AsyncMock()
        mock_res = MagicMock()
        MockWebSocketClient.return_value = mock_ws
        mock_ws.connect.return_value = None
        mock_ws.wait_until_identified.return_value = True
        mock_ws.call.return_value = mock_res
        mock_res.ok.return_value = False
        mock_res.requestType = 'SetInputSettings'
        mock_res.requestStatus.code = 600
        mock_res.requestStatus.comment = f'No source was found by the name of `{cfg_src_name}`.'

        obj = obsws.ObsWsEmitter(ctx=ctx, cfg=cfg)
        ctx.get_instance.assert_called_once_with(name=cfg_src)

        obj.logger.warning = MagicMock()

        await obj.update('test text', args=None)

        MockWebSocketClient.assert_called_with(url=cfg_url, password=cfg_password)
        mock_ws.connect.assert_awaited()
        mock_ws.call.assert_awaited_with(obsws.simpleobsws.Request(
            'SetInputSettings',
            {
                'inputName': cfg_src_name,
                'inputSettings': {'text': 'test text'},
            }
        ))
        self._assert_logging(obj.logger.warning, (
            ('%s: %d: %s', 'SetInputSettings', 600, f'No source was found by the name of `{cfg_src_name}`.'),
            ('%s: %d: %s', 'SetInputSettings', 600, f'No source was found by the name of `{cfg_src_name}`.'),
        ))


    def _assert_logging(self, mock_logging, expects):
        for ix, exp in enumerate(expects):
            for j, e in enumerate(exp):
                self.assertEqual(str(mock_logging.call_args_list[ix][0][j]), str(e))
        self.assertEqual(len(mock_logging.call_args_list), len(expects))

if __name__ == '__main__':
    unittest.main()



import sys
import tempfile
import unittest
from unittest.mock import MagicMock, patch

from slidetextbridge.core import main

@unittest.skipUnless(sys.version_info >= (3, 12), 'Requires Python >= 3.12')
class TestMainFailure(unittest.TestCase):

    @patch('logging.getLogger')
    def test_empty_config(self, mock_logger):
        with tempfile.NamedTemporaryFile(mode='w', delete_on_close=False) as fc:
            fc.write('steps:')
            fc.close()

            mock_logger.return_value = MagicMock()
            mock_error = mock_logger.return_value.error

            with patch('sys.argv', ['test', '-c', fc.name]):
                ret = main.main()

            self.assertNotEqual(ret, 0)
            # TODO: Better message
            self.assertIn('Failed to start', mock_error.call_args_list[0][0][0])

    @patch('logging.getLogger')
    def test_src_connection_fail(self, mock_logger):
        with tempfile.NamedTemporaryFile(mode='w', delete_on_close=False) as fc:
            fc.write('''
            steps:
              - type: stdout
             ''')
            fc.close()

            mock_logger.return_value = MagicMock()
            mock_error = mock_logger.return_value.error

            with patch('sys.argv', ['test', '-c', fc.name]):
                ret = main.main()

            self.assertNotEqual(ret, 0)
            self.assertIn('Cannot connect to the source', mock_error.call_args_list[0][0][0])

    @patch('logging.getLogger')
    def test_no_type(self, mock_logger):
        with tempfile.NamedTemporaryFile(mode='w', delete_on_close=False) as fc:
            fc.write('''
            steps:
              - json: true
             ''')
            fc.close()

            mock_logger.return_value = MagicMock()
            mock_error = mock_logger.return_value.error

            with patch('sys.argv', ['test', '-c', fc.name]):
                ret = main.main()

            self.assertNotEqual(ret, 0)
            self.assertIn('A field "type" is missing.', mock_error.call_args_list[0][0][0])

    @patch('logging.getLogger')
    def test_wrong_type(self, mock_logger):
        with tempfile.NamedTemporaryFile(mode='w', delete_on_close=False) as fc:
            fc.write('''
            steps:
              - type: invalid-type
             ''')
            fc.close()

            mock_logger.return_value = MagicMock()
            mock_error = mock_logger.return_value.error

            with patch('sys.argv', ['test', '-c', fc.name]):
                ret = main.main()

            self.assertNotEqual(ret, 0)
            self.assertIn('No such type', mock_error.call_args_list[0][0][0])

    @patch('logging.getLogger')
    def test_wrong_step_field(self, mock_logger):
        with tempfile.NamedTemporaryFile(mode='w', delete_on_close=False) as fc:
            fc.write('''
            steps:
              - type: openlp
                invalid_field: 'error'
             ''')
            fc.close()

            mock_logger.return_value = MagicMock()
            mock_error = mock_logger.return_value.error

            with patch('sys.argv', ['test', '-c', fc.name]):
                ret = main.main()

            self.assertNotEqual(ret, 0)
            self.assertIn('Failed to start', mock_error.call_args_list[0][0][0])
            self.assertIn('is undefined', str(mock_error.call_args_list[0][0][1]))

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock, AsyncMock

from slidetextbridge.core import context

class TestContext(unittest.IsolatedAsyncioTestCase):

    def test_instances(self):

        ctx = context.Context()

        inst1 = MagicMock()
        inst1.name = 'name1'
        inst2 = MagicMock()
        inst2.name = 'name2'

        ctx.add_instance(inst1)

        self.assertEqual(ctx.get_instance(), inst1)

        ctx.add_instance(inst2)

        self.assertEqual(ctx.get_instance('name1'), inst1)
        self.assertEqual(ctx.get_instance('name2'), inst2)
        self.assertEqual(ctx.get_instance(), inst2)

    async def test_init(self):

        ctx = context.Context()

        inst1 = AsyncMock()
        inst2 = AsyncMock()

        ctx.add_instance(inst1)
        ctx.add_instance(inst2)

        await ctx.initialize_all()

        inst1.initialize.assert_awaited_once()
        inst2.initialize.assert_awaited_once()

if __name__ == '__main__':
    unittest.main()



'''
Filters to modify texts
'''

import re
import logging
from slidetextbridge.core import config
from . import base


_CJK_RANGES = (
          (0x03300, 0x033FF),  # compatibility ideographs
          (0x0FE30, 0x0FE4F),  # compatibility ideographs
          (0x0F900, 0x0FAFF),  # compatibility ideographs
          (0x2F800, 0x2FA1F),  # compatibility ideographs
          (0x03000, 0x0303F),  # CJK symbols and punctuation
          (0x03040, 0x030FF),  # Japanese Hiragana and Katakana
          (0x03190, 0x0319F),  # Kanbun
          (0x02E80, 0x02EFF),  # CJK radicals supplement
          (0x04E00, 0x09FFF),
          (0x03400, 0x04DBF),
          (0x0AC00, 0x0D7AF),  # Hangul Syllables
          (0x1B130, 0x1B16F),  # Small Kana Extension
          (0x20000, 0x2A6DF),
          (0x2A700, 0x2B73F),
          (0x2B740, 0x2B81F),
          (0x2B820, 0x2CEAF),  # included as of Unicode 8.0
)

def _is_cjk_char(c):
    c = ord(c)
    for r in _CJK_RANGES:
        if r[0] <= c <= r[1]:
            return True
    return False

class TextLinebreakFilter(base.PluginBase):
    '''
    Filter line breaks in texts
    '''

    @staticmethod
    def type_name():
        return 'linebreak'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('shape_delimiter', type=str, default='\n')
        cfg.add_argment('line_delimiter', type=str, default='\n')
        cfg.add_argment('strip', type=bool, default=True)
        cfg.add_argment('split_long_line', type=int, default=0)
        cfg.add_argment('split_nowrap', type=str, default='.,"\'')
        cfg.add_argment('split_nowrap_allow_overflow', type=bool, default=True)
        cfg.add_argment('joined_column_max', type=int, default=0)
        cfg.add_argment('join_by', type=str, default=' ')
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'linebreak({self.cfg.location})')
        self.connect_to(cfg.src)

    def _filter_shape_text(self, text):
        lines = text.split('\n')

        if self.cfg.strip:
            stripped = []
            for t in lines:
                t = t.strip()
                if t:
                    stripped.append(t)
            lines = stripped

        if self.cfg.split_long_line:
            new_lines = []
            for line in lines:
                new_lines += self._split_long_line(line)
            lines = new_lines

        if self.cfg.joined_column_max:
            joined = []
            pending_text = None
            for t in lines:
                next_text = pending_text + self.cfg.join_by + t if pending_text else t
                nt = self._count_text(next_text)
                if nt > self.cfg.joined_column_max:
                    if pending_text:
                        joined.append(pending_text)
                    pending_text = t
                else:
                    pending_text = next_text
            if pending_text:
                joined.append(pending_text)
            lines = joined

        return self.cfg.line_delimiter.join(lines)

    def _split_long_line(self, text):
        # pylint: disable=R0912
        if self._count_text(text) <= self.cfg.split_long_line:
            yield text
            return

        while text:
            # Find the length that satisfies `split_long_line`
            ix = 0
            ix_cjk = -1
            ix_space = -1
            ix_other = -1
            n = 0
            allow_overflow = False
            while ix < len(text):
                c = text[ix]
                n += self._count_text(c)
                allow_overflow = False
                if c.isspace():
                    ix_space = ix
                elif self.cfg.split_nowrap and c in self.cfg.split_nowrap:
                    if self.cfg.split_nowrap_allow_overflow:
                        allow_overflow = True
                elif _is_cjk_char(c):
                    ix_cjk = ix
                else:
                    ix_other = ix

                if not allow_overflow and n > self.cfg.split_long_line:
                    break
                ix += 1

            if ix == len(text):
                yield text
                break

            # If there is a space, let's cut at the space.
            if ix_space > 0 and ix_space > ix_cjk:
                yield text[:ix_space]
                text = text[ix_space+1:]

            # If there is a CJK string,
            elif ix_cjk > 0:
                yield text[:ix_cjk]
                text = text[ix_cjk:]

            # If no more option to break
            elif ix_other > 0:
                yield text[:ix_other]
                text = text[ix_other:]

            # The last condition to avoid infinite loop
            # Will reach when satisfying both of these conditions.
            # - There are only nowrap characters but exceeding the threshold.
            # - allow_overflow = false.
            else:
                yield text
                break

    def _count_text(self, text):
        'Count CJK characters twice'
        if text.isascii():
            return len(text)
        n = 0
        for c in text:
            n += 2 if _is_cjk_char(c) else 1
        return n

    async def update(self, slide, args):
        texts = slide.to_texts()
        shapes = [{
            'text': self._filter_shape_text(t),
            'shape_delimiter': self.cfg.shape_delimiter
            } for t in texts]
        slide = TextFilteredSlide(data={'shapes': shapes}, parent=self)
        await self.emit(slide)


def _parse_patterns(v):
    cfg = config.ConfigBase()
    cfg.add_argment('p', type=str)
    cfg.add_argment('r', type=str)
    ret = []
    for pattern in v:
        cfg.parse(pattern)
        ret.append((re.compile(cfg.p), cfg.r)) # type: ignore[attr-defined] # pylint: disable=E1101

    return ret

class RegexFilter(base.PluginBase):
    '''
    Filter lines with regex
    '''

    @classmethod
    def type_name(cls):
        return 'regex'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('patterns', conversion=_parse_patterns)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'regex({self.cfg.location})')
        self.connect_to(cfg.src)

    def _filter_shape_text(self, text):
        lines = text.split('\n')
        for p, r in self.cfg.patterns:
            lines = [p.sub(r, t) for t in lines]
        return '\n'.join(lines)

    async def update(self, slide, args):
        texts = slide.to_texts()
        shapes = [{
            'text': self._filter_shape_text(t),
            } for t in texts]
        slide = TextFilteredSlide(data={'shapes': shapes}, parent=self)
        await self.emit(slide)


class TextFilteredSlide(base.SlideBase):
    'The slide class returned by the filters'

    def __init__(self, data=None, parent=None):
        if isinstance(data, list):
            self._dict = {'shapes': data}
        else:
            self._dict = data
        self.parent = parent

    def to_texts(self):
        try:
            if self._dict:
                return [shape['text'] for shape in self._dict['shapes']]
        except (TypeError, KeyError) as e:
            logger = self.parent.logger if self.parent else logging.getLogger('TextFilteredSlide')
            logger.error('Failed to convert slide to texts. %s', e)
        return []

    def __str__(self):
        ret = ''
        shape_delimiter = ''
        for shape in self._dict['shapes']:
            ret += shape_delimiter
            ret += shape['text']
            shape_delimiter = shape['shape_delimiter'] if 'shape_delimiter' in shape else '\n'
        return ret

    def to_dict(self):
        return self._dict



import os
import sys
import types
import unittest
from unittest.mock import MagicMock, patch, AsyncMock

mock_win32com_client = types.SimpleNamespace(Dispatch=lambda x: None)
mock_win32com = types.SimpleNamespace(client=mock_win32com_client)
mock_pywintypes = types.SimpleNamespace(com_error=Exception)
sys.modules['win32com'] = mock_win32com
sys.modules['win32com.client'] = mock_win32com_client
sys.modules['pywintypes'] = mock_pywintypes

from slidetextbridge.plugins import powerpoint # pylint: disable=C0413

def mock_shape(text, shape_type=14, name='unnamed'):
    shape = MagicMock()

    shape.HasTextFrame = True
    shape.TextFrame = MagicMock()
    shape.TextFrame.TextRange = MagicMock()
    shape.TextFrame.TextRange.Text = text
    shape.TextFrame.TextRange.HasText = len(text) > 0
    shape.TextFrame.TextRange.Count = len(text)
    shape.TextFrame.TextRange.Start = 0
    shape.TextFrame.TextRange.Length = len(text)
    shape.TextFrame.TextRange.BoundLeft = 0
    shape.TextFrame.TextRange.BoundTop = 0
    shape.TextFrame.TextRange.BoundWidth = 1
    shape.TextFrame.TextRange.BoundHeight = 1
    shape.TextFrame.TextRange.Font = MagicMock()
    shape.TextFrame.TextRange.Font.Size = 24
    shape.TextFrame.TextRange.Font.Bold = False
    shape.TextFrame.TextRange.Font.Name = name
    shape.TextFrame.TextRange.Font.BaselineOffset = 0
    shape.TextFrame.TextRange.Font.Italic = False
    shape.TextFrame.TextRange.Font.Subscript = False
    shape.TextFrame.TextRange.Font.Superscript = False
    shape.TextFrame.HasText = True
    shape.TextFrame.Orientation = 0
    shape.TextFrame.WordWrap = False

    shape.PlaceholderFormat = MagicMock()
    shape.PlaceholderFormat.Name = name
    shape.PlaceholderFormat.Type = 1
    shape.PlaceholderFormat.ContainedType = 2

    shape.Type = shape_type
    shape.Name = name

    return shape

class TestPowerPointCapture(unittest.IsolatedAsyncioTestCase):
    def test_type_name(self):
        self.assertEqual(powerpoint.PowerPointCapture.type_name(), 'ppt')

    async def test_loop(self):
        ctx = MagicMock()
        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        obj._loop_once = AsyncMock()
        obj._loop_once.side_effect = (None, )
        obj.logger = MagicMock()
        with patch('asyncio.sleep', side_effect=(None, )) as mock_sleep:
            with self.assertRaises(StopAsyncIteration):
                await obj._loop()

        self.assertEqual(obj._loop_once.await_count, 3)
        self.assertEqual(mock_sleep.await_count, 2)

    @patch('win32com.client.Dispatch', autospec=True)
    async def test_update(self, MockDispatch):
        ctx = MagicMock()

        ppt = MagicMock()
        MockDispatch.return_value = ppt

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        win = MagicMock()
        api_slide = MagicMock()
        ppt.SlideShowWindows.Count = 1
        ppt.SlideShowWindows.return_value = win
        win.View = MagicMock()
        win.View.State = 0
        win.View.Slide = api_slide

        api_slide.Shapes = [
                mock_shape(text='a'),
                mock_shape(text='b'),
        ]
        api_slide.Shapes[1].Type = 13 # Not a placeholder

        obj.emit = AsyncMock()

        await obj._loop_once()

        obj.emit.assert_called_once()
        slide = obj.emit.call_args[0][0]

        MockDispatch.assert_called_once_with('PowerPoint.Application')

        self.assertEqual(str(slide), 'a')
        # self.maxDiff = None
        d = slide.to_dict()
        self.assertEqual(d['shapes'][0]['text_frame']['has_text'], True)
        self.assertEqual(d['shapes'][0]['text_frame']['text_range']['text'], 'a')
        # TODO: Check other fields. I should copy the expected data from actual PowerPoint.

        self.assertEqual(slide.to_texts(), ['a'])

        # Call again, Dispatch should not be called again.
        await obj._loop_once()
        MockDispatch.assert_called_once_with('PowerPoint.Application')

    @patch('win32com.client.Dispatch', autospec=True)
    async def test_multi_windows(self, MockDispatch):
        ctx = MagicMock()

        ppt = MagicMock()
        MockDispatch.return_value = ppt

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        win1 = MagicMock()
        api_slide = MagicMock()
        win1.View = MagicMock()
        win1.View.State = 0
        win1.View.Slide = api_slide
        win1.Active = False

        api_slide.Shapes = [
                mock_shape(text='win1-shape'),
        ]

        win2 = MagicMock()
        api_slide = MagicMock()
        win2.View = MagicMock()
        win2.View.State = 0
        win2.View.Slide = api_slide
        win2.Active = True

        api_slide.Shapes = [
                mock_shape(text='win2-shape'),
        ]

        ppt.SlideShowWindows.Count = 2
        ppt.SlideShowWindows.side_effect = (win1, win2)

        obj.emit = AsyncMock()

        await obj._loop_once()

        obj.emit.assert_called_once()
        slide = obj.emit.call_args[0][0]

        self.assertEqual(str(slide), 'win2-shape')

        ppt.SlideShowWindows.Count = 2
        ppt.SlideShowWindows.side_effect = (win1, win2)
        api_slide = MagicMock()
        win2.Active = False
        win2.View.Slide = api_slide
        api_slide.Shapes = [
                mock_shape(text='win2-shape modified'),
        ]

        obj.emit = AsyncMock()
        await obj._loop_once()
        obj.emit.assert_called_once()
        slide = obj.emit.call_args[0][0]
        self.assertEqual(str(slide), 'win2-shape modified')

    @patch('win32com.client.Dispatch', autospec=True)
    async def test_dispatch_failure(self, MockDispatch):
        ctx = MagicMock()

        MockDispatch.side_effect = mock_pywintypes.com_error
        MockDispatch.return_value = None

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        obj.emit = AsyncMock()

        await obj._loop_once()

        self._assert_empty_slide(obj.emit)

    @patch('win32com.client.Dispatch', autospec=True)
    async def test_ppt_windows_count_exception(self, MockDispatch):
        ctx = MagicMock()

        ppt = MagicMock()
        MockDispatch.return_value = ppt

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        def _raise_error(*args, **kwargs):
            raise mock_pywintypes.com_error
        type(ppt.SlideShowWindows).Count = property(_raise_error)

        obj.emit = AsyncMock()

        await obj._loop_once()

        self._assert_empty_slide(obj.emit)

    @patch('win32com.client.Dispatch', autospec=True)
    async def test_ppt_windows_count_0(self, MockDispatch):
        ctx = MagicMock()

        ppt = MagicMock()
        MockDispatch.return_value = ppt

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        ppt.SlideShowWindows.Count = 0

        obj.emit = AsyncMock()

        await obj._loop_once()

        self._assert_empty_slide(obj.emit)

    def _assert_empty_slide(self, mock_emit):
        mock_emit.assert_called_once()
        slide = mock_emit.call_args[0][0]
        self.assertEqual(slide._slide, None)
        self.assertEqual(slide._dict, None)

    @patch('win32com.client.Dispatch', autospec=True)
    def test_blank(self, MockDispatch):
        ctx = MagicMock()

        ppt = MagicMock()
        MockDispatch.return_value = ppt

        cfg = powerpoint.PowerPointCapture.config({})
        obj = powerpoint.PowerPointCapture(ctx=ctx, cfg=cfg)

        win = MagicMock()
        ppt.SlideShowWindows.Count = 1
        ppt.SlideShowWindows.return_value = win
        win.View = MagicMock()
        win.View.State = powerpoint._Const.ppSlideShowBlackScreen

        int_slide = obj._get_slide()

        self.assertEqual(int_slide, None)

    def test_accumulate_powerpoint(self):
        import importlib
        with patch('os.name', 'nt-test'):
            accumulate_name = 'slidetextbridge.plugins.accumulate'
            if accumulate_name in sys.modules:
                importlib.reload(sys.modules[accumulate_name])
            else:
                importlib.import_module(accumulate_name)
            accumulate = sys.modules[accumulate_name]
            self.assertEqual(accumulate._is_win, True)
            self.assertIn('ppt', accumulate.plugins)


if __name__ == "__main__":
    unittest.main()



'''
Get text from OpenLP
'''

import json
import asyncio
import logging
import aiohttp
import aiohttp.client_exceptions
import websockets
from slidetextbridge.core import config
from slidetextbridge.core.logging import HideSameLogFilter
from . import base


def _is_blank(poll):
    if poll['results']['blank']:
        return True
    if poll['results']['theme']:
        return True
    if poll['results']['display']:
        return True
    return False

def _filter_by_types(slide, item):
    item_type = item['type']
    if item_type == 1: # bibles songs custom
        return slide
    return None


class OpenLPCapture(base.PluginBase):
    '''
    Get text from OpenLP
    '''
    @staticmethod
    def type_name():
        return 'openlp'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg, has_src=False)
        cfg.add_argment('host', type=str, default='localhost')
        cfg.add_argment('port', type=int, default=4316)
        cfg.add_argment('port_ws', type=int, default=None)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'openlp({self.cfg.location})')
        self.logger.addFilter(HideSameLogFilter())

        self._conn = None
        self._conn_ws = None
        self._port_ws = None
        self._cache = {}
        if not self.cfg.port_ws:
            self.cfg.port_ws = self.cfg.port + 1

    async def _connect(self):
        if not self._conn:
            self._conn = aiohttp.TCPConnector()

    async def _connect_ws(self):
        url = f'ws://{self.cfg.host}:{self.cfg.port_ws:d}/poll'
        try:
            return await websockets.connect(url)
        except OSError as e:
            raise OSError(f'Failed to connect to {url}. {str(e)}') from e

    async def _olp_get(self, url):
        if not self._conn:
            await self._connect()
        if url[0] == '/':
            url = f'http://{self.cfg.host}:{self.cfg.port}{url}'
        async with aiohttp.request('GET', url, connector=self._conn) as res:
            txt = await res.text()
            return txt

    async def _olp_poll(self):
        if not self._conn_ws:
            self._conn_ws = await self._connect_ws()
        try:
            res = await self._conn_ws.recv()
            return json.loads(res)
        except Exception as e: # pylint: disable=W0718
            await self._conn_ws.close()
            self._conn_ws = None
            self.logger.warning('Failed to poll. %s', e)
            return None

    async def _cache_current_item(self):
        live_items = json.loads(await self._olp_get('/api/v2/controller/live-items'))
        self._cache[live_items['id']] = live_items

    async def _wait_and_get_text(self):
        while True:
            poll = await self._olp_poll()
            if not poll:
                return None
            if _is_blank(poll):
                return None
            poll_res = poll['results']
            item_uuid = poll_res['item']
            if item_uuid not in self._cache:
                try:
                    await self._cache_current_item()
                except aiohttp.client_exceptions.ClientOSError as e:
                    self.logger.warning('Failed to get live-items. %s', e)
                    return None
            if item_uuid in self._cache:
                current_item = self._cache[item_uuid]
                break

        i_slide = poll_res['slide']
        return _filter_by_types(current_item['slides'][i_slide], current_item)

    async def _loop_once(self):
        obj = await self._wait_and_get_text()
        slide = OpenLPSlide(data = {'shapes': [obj] if obj else []}, parent=self)
        await self.emit(slide)

    async def initialize(self):
        asyncio.create_task(self._loop())

    async def _loop(self):
        while True:
            try:
                await self._loop_once()
            except Exception as e:
                self.logger.warning('%s', e)
                await asyncio.sleep(3)
            if not self._conn_ws:
                await asyncio.sleep(3)


class OpenLPSlide(base.SlideBase):
    'The slide class returned by OpenLPCapture'

    def __init__(self, data=None, parent=None):
        if isinstance(data, list):
            self._dict = {'shapes': data}
        else:
            self._dict = data
        self.parent = parent

    def to_texts(self):
        try:
            if self._dict:
                return [shape['text'] for shape in self._dict['shapes']]
        except (TypeError, KeyError) as e:
            logger = self.parent.logger if self.parent else logging.getLogger('OpenLPSlide')
            logger.error('Failed to convert slide to texts. %s', e)
        return []

    def to_dict(self):
        return self._dict



import unittest
from unittest.mock import MagicMock, patch, AsyncMock
from aiohttp.test_utils import AioHTTPTestCase

from slidetextbridge.plugins import webserver

# https://docs.aiohttp.org/en/stable/testing.html#unittest

class TestWebServerEmitter(AioHTTPTestCase):

    def test_type_name(self):
        self.assertEqual(webserver.WebServerEmitter.type_name(), 'webserver')

    def test_config(self):
        cfg_data = {'host': '192.0.2.3', 'port': 8081}
        cfg = webserver.WebServerEmitter.config(cfg_data)
        self.assertEqual(cfg.host, '192.0.2.3')
        self.assertEqual(cfg.port, 8081)

    async def get_application(self):
        ctx = MagicMock()

        cfg_data = {
                'host': '192.0.2.3',
                'port': 8081,
                'index_html': '<!-- index html -->',
                'script_js': '/* script data */',
                'style_css': '/* style data */',
        }
        cfg = webserver.WebServerEmitter.config(cfg_data)

        mockAppRunner = MagicMock()
        mockAppRunner.return_value = MagicMock()
        mockAppRunner.return_value.setup = AsyncMock()
        mockTCPSite = MagicMock()
        mockTCPSite.return_value = MagicMock()
        mockTCPSite.return_value.start = AsyncMock()

        with patch.multiple(
                'aiohttp.web',
                AppRunner=mockAppRunner,
                TCPSite=mockTCPSite,
        ):
            obj = webserver.WebServerEmitter(ctx=ctx, cfg=cfg)
            await obj.initialize()

        mockAppRunner.assert_called_once_with(obj.app)
        mockAppRunner.return_value.setup.assert_called_once_with()
        mockTCPSite.assert_called_once_with(
                mockAppRunner.return_value,
                obj.cfg.host, obj.cfg.port)
        mockTCPSite.return_value.start.assert_called_once_with()

        self.obj = obj
        return obj.app

    async def test_get(self):

        async with self.client.request('GET', '/') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '<!-- index html -->')

        async with self.client.request('GET', '/script.js') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '/* script data */')

        async with self.client.request('GET', '/style.css') as resp:
            self.assertEqual(resp.status, 200)
            text = await resp.text()
            self.assertEqual(text, '/* style data */')

    async def test_ws_text(self):

        # no text are pending
        async with self.client.ws_connect('/ws/text') as ws:
            await self.obj.update('first text', None)
            self.assertEqual(await ws.receive_str(), 'first text')

        # pending text will be sent soon
        await self.obj.update('pending text', None)
        async with self.client.ws_connect('/ws/text') as ws:
            await ws.send_str('ignored text')
            await self.obj.update('updated text', None)
            self.assertEqual(await ws.receive_str(), 'pending text')
            self.assertEqual(await ws.receive_str(), 'updated text')

if __name__ == '__main__':
    unittest.main()



'''
Send text through HTTP
'''

import logging
from aiohttp import web
from slidetextbridge.core import config
from . import base

_DEFAULT_INDEX_HTML = '''
<!doctype html>
<html>
    <header>
        <script src="{uri_script}"></script>
        <link rel='stylesheet' href='{uri_style}' />
    </header>
    <body>
        <div class='text' id='placeholder'>
        </div>
    </body>
</html>
'''.strip()

_DEFAULT_SCRIPT_JS = '''
function get_appropriate_ws_url(extra_url) {
    var pcol;
    var u = document.URL;
    if (u.substring(0, 5) === "https") {
        pcol = "wss://";
        u = u.substr(8).split("/")[0];
    } else {
        pcol = "ws://";
        if (u.substring(0, 4) === "http")
            u = u.substr(7).split("/")[0];
    }
    return pcol + u + "/" + extra_url;
}

function text_to_html(text) {
    return text.replace(/[&<>"']/g, function (c) {
        return ({
            '&': '&amp;',
            '<': '&lt;',
            '>': '&gt;',
            '"': '&quot;',
            "'": '&#39;',
        })[c];
    }).replace(/\\n/g, "<br/>");
}

let ws_failed = 0;
function connect_ws() {
    let url = get_appropriate_ws_url("ws/text");
    console.log("Connecting to", url);
    ws = new WebSocket(url);
    ws.onmessage = function(msg) {
        ws_failed = 0;
        console.log("Received:", msg.data);
        e = document.getElementById("placeholder");
        h = text_to_html(msg.data);
        console.log("HTML:", h);
        e.innerHTML = h;
    };
    ws.onclose = function() {
        ws_failed += 1;
        setTimeout(connect_ws, ws_failed < 6 ? ws_failed * 500 : 3000);
    }
}

document.addEventListener("DOMContentLoaded", function() {
    connect_ws();
});
'''

_DEFAULT_STYLE_CSS = '''
body {
  font-family: sans-serif;
}
'''

class WebServerEmitter(base.PluginBase):
    '''
    HTTP server to send text through HTTP
    '''
    @staticmethod
    def type_name():
        return 'webserver'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('host', type=str, default='localhost')
        cfg.add_argment('port', type=int, default=8080)
        cfg.add_argment('uri_index', type=str, default='/index.html')
        cfg.add_argment('uri_script', type=str, default='/script.js')
        cfg.add_argment('uri_style', type=str, default='/style.css')
        cfg.add_argment('index_html', type=str, default=_DEFAULT_INDEX_HTML)
        cfg.add_argment('script_js', type=str, default=_DEFAULT_SCRIPT_JS)
        cfg.add_argment('style_css', type=str, default=_DEFAULT_STYLE_CSS)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'webserver({self.cfg.location})')
        self.app = None
        self.clients = set()
        self._last_text = None
        self.connect_to(cfg.src)

    def _fmt(self, text):
        return text \
                .replace('{uri_script}', self.cfg.uri_script) \
                .replace('{uri_style}', self.cfg.uri_style)

    async def _handle_index(self, request):
        # pylint: disable=unused-argument
        return web.Response(text=self._fmt(self.cfg.index_html), content_type='text/html')

    async def _handle_script(self, request):
        # pylint: disable=unused-argument
        return web.Response(text=self._fmt(self.cfg.script_js), content_type='text/javascript')

    async def _handle_style(self, request):
        # pylint: disable=unused-argument
        return web.Response(text=self._fmt(self.cfg.style_css), content_type='text/css')

    async def _handle_ws(self, request):
        # pylint: disable=unused-argument
        self.logger.info('Starting %s remote=%s', request.rel_url, request.remote)
        try:
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            self.clients.add(ws)

            if self._last_text:
                await self._send_text(ws, self._last_text)

            async for _ in ws:
                pass
        except Exception as e:
            self.logger.error('%s: Failed to process ws message. %s', request.remote, e)
        finally:
            self.clients.discard(ws)

        self.logger.info('Closing %s remote=%s', request.rel_url, request.remote)
        return ws

    async def initialize(self):
        self.app = web.Application()
        self.app.add_routes([
            web.get('/', self._handle_index),
            web.get(self.cfg.uri_index, self._handle_index),
            web.get(self.cfg.uri_script, self._handle_script),
            web.get(self.cfg.uri_style, self._handle_style),
            web.get('/ws/text', self._handle_ws),
        ])
        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, self.cfg.host, self.cfg.port)
        await site.start()
        self.logger.info('Listening on %s:%d', self.cfg.host, self.cfg.port)

    async def update(self, slide, args):
        if not slide:
            text = ''
        elif isinstance(slide, str):
            text = slide
        else:
            text = str(slide)

        self._last_text = text
        for ws in self.clients:
            await self._send_text(ws, text)

    async def _send_text(self, ws, text):
        try:
            await ws.send_str(text)
        except Exception as e:
            self.logger.warning('Failed to send text to ws. %s', e)



'''
Configuration data structure
'''

_LIB_KEYS = ('_location', )

class ConfigBase():
    '''
    Base class to hold configuration data
    '''
    def __init__(self):
        self.argtypes = []
        self.location = None

    def add_argment(self, *args, **kwargs):
        '''
        Add an argument for this configuration type.
        Call this method before `parse`.
        :param name:       name of the argument (`snake_case` is recommended)
        :param type:       type of the argument
        :param conversion: conversion function taking one argument
        :param default:    the default value
        '''
        class _Argment:
            # pylint: disable=W0622,R0903
            def __init__(self, name, type=None, conversion=None, default=None):
                self.name = name
                self.type = type
                self.conversion = conversion
                self.default = default
        self.argtypes.append(_Argment(*args, **kwargs))

    def parse(self, d: dict):
        '''
        Parse the dictionary data and set attributes
        :param d:  data
        '''
        self.location = d['_location'] if '_location' in d else None
        for key in d.keys():
            if key in _LIB_KEYS:
                continue
            found = False
            for a in self.argtypes:
                if a.name == key:
                    found = True
                    break
            if not found:
                raise KeyError(f'{self.location}: {key} is undefined')
        for a in self.argtypes:
            if a.name in d:
                v = d[a.name]
                if a.conversion:
                    v = a.conversion(v)
                elif a.type:
                    v = a.type(v)
            else:
                v = a.default
            setattr(self, a.name, v)



'''
Main routine
'''

import argparse
import asyncio
import logging
import sys

from slidetextbridge.core.context import Context
from slidetextbridge.core import configtop

def _get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config', action='store', default='config.yaml')
    parser.add_argument('--strict', action='store_true')
    return parser.parse_args()

def _setup_ctx(cfgs):
    ctx = Context()
    from slidetextbridge.plugins import accumulate # pylint: disable=C0415
    for step in cfgs.steps:
        cls = accumulate.plugins[step.type]
        inst = cls(ctx=ctx, cfg=step)
        ctx.add_instance(inst)
    return ctx

async def _loop(ctx):
    await ctx.initialize_all()
    while True:
        tt = list(asyncio.all_tasks())
        t = asyncio.current_task()
        if t:
            tt.remove(t)
        if not tt:
            break
        rr = await asyncio.gather(*tt, return_exceptions=True)
        for t, r in zip(tt, rr):
            if isinstance(r, Exception):
                logging.getLogger(__name__).error('%s: Unknown exception %s', t, r)
    logging.getLogger(__name__).info('All tasks are done.')

def main():
    'The entry point'
    try:
        logging.basicConfig(level=logging.INFO)
        args = _get_args()
        cfgs = configtop.load(args.config)
        ctx = _setup_ctx(cfgs)
    except Exception as e:
        logging.getLogger(__name__).error('Failed to start. %s', e)
        return 1

    try:
        asyncio.run(_loop(ctx))
    except KeyboardInterrupt:
        logging.getLogger(__name__).info('Interrupted')

    return 0

if __name__ == '__main__':
    sys.exit(main())



'''
Filter with JMESPath
'''

import jmespath
from slidetextbridge.core import config
from . import base


class JMESPathFilter(base.PluginBase):
    '''
    Filter shapes with JMESPath
    '''
    @staticmethod
    def type_name():
        return 'jmespath'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('filter', type=str)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.connect_to(cfg.src)
        self.jmespath_filter = jmespath.compile(cfg.filter)

    async def update(self, slide, args):
        obj = slide.to_dict()
        obj = self.jmespath_filter.search(obj)
        slide = slide.__class__(data=obj)
        await self.emit(slide)



'''
Context for the program
'''

class Context:
    'Context class managing plugin instances'

    def __init__(self):
        self.instances = []

    def add_instance(self, inst):
        '''
        Add an instance of a plugin
        :param inst:  The instance to be added
        '''
        self.instances.append(inst)

    def get_instance(self, name=None):
        '''
        Find instance
        If no argument is specified, return the last
        :param name:  If given, find by name
        '''
        if name:
            for inst in self.instances:
                if inst.name == name:
                    return inst
            raise KeyError(name)
        return self.instances[-1]

    async def initialize_all(self):
        '''
        After all the instances are registered, call this method to initialize them.
        '''
        for inst in self.instances:
            await inst.initialize()



import unittest
from unittest.mock import MagicMock, AsyncMock

from slidetextbridge.plugins.jmespath_filter import JMESPathFilter
from slidetextbridge.plugins.text_filters import TextFilteredSlide
from slidetextbridge.plugins import base

class TestJMESPathFilter(unittest.IsolatedAsyncioTestCase):

    def test_type_name(self):
        self.assertEqual(JMESPathFilter.type_name(), 'jmespath')

    def test_config(self):
        cfg_filter = 'filter-text'
        cfg = JMESPathFilter.config(
                data = {'filter': cfg_filter}
        )

        self.assertEqual(cfg.filter, cfg_filter)

    async def test_update_filters_slide(self):
        ctx = MagicMock()

        cfg = MagicMock()
        cfg.src = 'dummy'
        cfg.filter = 'shapes[?val==`2`]'

        filter_obj = JMESPathFilter(ctx=ctx, cfg=cfg)
        filter_obj.emit = AsyncMock()

        slide = TextFilteredSlide(data={'shapes': [{'text': 'a', 'val': 1}, {'text': 'b', 'val': 2}]})

        await filter_obj.update(slide, args=None)

        filter_obj.emit.assert_awaited_once()
        emitted_slide = filter_obj.emit.await_args[0][0]
        self.assertIsInstance(emitted_slide, TextFilteredSlide)
        self.assertEqual(emitted_slide.to_dict(), {'shapes': [{'text': 'b', 'val': 2}]})
        self.assertEqual(emitted_slide.to_texts(), ['b', ])

        # If nothing is filtered,
        slide = TextFilteredSlide(data={'shapes': [{'text': 'a', 'val': 1}, {'text': 'b', 'val': 3}]})
        await filter_obj.update(slide, args=None)
        emitted_slide = filter_obj.emit.await_args[0][0]
        self.assertEqual(emitted_slide.to_texts(), [])
        self.assertEqual(emitted_slide.to_dict(), {'shapes': []})

        # TextFilteredSlide ignores shapes that does not contain the `text` field.
        slide = TextFilteredSlide(data={'shapes': [{'test': 'a', 'val': 1}, {'test': 'b', 'val': 2}]})
        await filter_obj.update(slide, args=None)
        emitted_slide = filter_obj.emit.await_args[0][0]
        self.assertEqual(emitted_slide.to_texts(), [])

        # TextFilteredSlide ignores shapes list directly having string.
        slide = TextFilteredSlide(data={'shapes': [{'text': 'c', 'val': 1}, {'text': 'd', 'val': 2}]})
        cfg.filter = 'shapes[?val==`2`].text'
        filter_obj = JMESPathFilter(ctx=ctx, cfg=cfg)
        filter_obj.emit = AsyncMock()
        await filter_obj.update(slide, args=None)
        emitted_slide = filter_obj.emit.await_args[0][0]
        self.assertEqual(emitted_slide.to_dict(), {'shapes': ['d']})
        self.assertEqual(emitted_slide.to_texts(), [])

if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

import setuptools

with open('README.md', 'rt', encoding='ascii') as f:
    long_description = f.read()

short_description = long_description.split('\n', 1)[0].strip('# ')

with open('requirements.txt', 'rt', encoding='ascii') as f:
    requirements = f.readlines()

setuptools.setup(
        name = 'slidetextbridge',
        version='0.2.1',
        description=short_description,
        long_description=long_description,
        long_description_content_type='text/markdown',
        package_dir={'': 'src'},
        packages=setuptools.find_packages(where='src'),
        install_requires=requirements,
        python_requires='>=3.9',
        entry_points={
            'console_scripts': [
                'slidetextbridge=slidetextbridge.core.main:main',
            ],
        },
        url='https://github.com/norihiro/slidetextbridge',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Intended Audience :: End Users/Desktop',
            'Environment :: Console',
        ],
)



'''
Configuration file reader
'''

import logging
import yaml
from .config import ConfigBase

class ConfigTop(ConfigBase):
    # pylint: disable=R0903
    '''
    The class to hold every configuration
    '''
    def __init__(self):
        super().__init__()
        self.add_argment('steps', conversion=ConfigTop._steps_converter, default=[])

    @staticmethod
    def _steps_converter(data):
        from slidetextbridge.plugins import accumulate # pylint: disable=C0415
        ret = []
        for d in data:
            try:
                t = d['type']
            except KeyError as e:
                logging.getLogger(__name__).error('%s: A field "type" is missing.', d['_location'])
                raise ValueError(f'Error while parsing {d["_location"]}') from e
            try:
                cls = accumulate.plugins[t]
            except KeyError as e:
                logging.getLogger(__name__).error('%s: No such type %s', d['_location'], e)
                raise ValueError(f'Error while parsing {d["_location"]}') from e
            cfg = cls.config(d)
            ret.append(cfg)
        return ret


def load(filename):
    '''
    Load the config file in YAML
    :param filename: The file name to read
    '''

    class _LoaderWithLineNumber(yaml.SafeLoader): # pylint: disable=R0901
        _filename = filename
        def construct_mapping(self, node, deep=False):
            mapping = super().construct_mapping(node, deep=deep)
            location = f'{self._filename}:{node.start_mark.line+1}'
            if node.start_mark.line != node.end_mark.line:
                location += f'-{node.end_mark.line+1}'
            mapping['_location'] = location
            return mapping

    with open(filename, 'r', encoding='utf8') as fr:
        d = yaml.load(fr, Loader=_LoaderWithLineNumber)

    cfgs = ConfigTop()
    cfgs.parse(d)
    return cfgs



'''
Helper classes for debugging
'''

import json
import logging
import sys
from slidetextbridge.core import config
from . import base


class StdoutEmitter(base.PluginBase):
    '''
    Just print to stdout
    '''
    @staticmethod
    def type_name():
        return 'stdout'

    @staticmethod
    def config(data):
        'Return the config object'
        cfg = config.ConfigBase()
        base.set_config_arguments(cfg)
        cfg.add_argment('page_delimiter', type=str, default='\n\n')
        cfg.add_argment('json', type=bool, default=False)
        cfg.parse(data)
        return cfg

    def __init__(self, ctx, cfg=None):
        super().__init__(ctx=ctx, cfg=cfg)
        self.logger = logging.getLogger(f'stdout({self.cfg.location})')
        self.connect_to(cfg.src)

    async def update(self, slide, args):
        try:
            if self.cfg.json:
                text = json.dumps(slide.to_dict(), ensure_ascii=False, indent=2)
            elif not slide:
                text = ''
            else:
                text = str(slide)

            sys.stdout.write(f'{text}{self.cfg.page_delimiter}')
        except Exception as e:
            self.logger.error('%s', e)

        await self.emit(slide)



'''
Helper classes and functions to log messages
'''

import logging

class HideSameLogFilter(logging.Filter):
    # pylint: disable=R0903
    '''
    Filter class for logger that removes same log messages
    More precisely, this class saves recent log messages limited by `limit`,
    compares the new message with the saved log messages,
    discard the message if matched.
    '''
    def __init__(self, limit=3):
        super().__init__()
        self._lasts = []
        self._limit = limit

    def filter(self, record):
        for last in self._lasts:
            if last == record.msg:
                return False

        if len(self._lasts) == self._limit:
            self._lasts = self._lasts[1:]
        self._lasts.append(record.msg)

        return True



import json
import unittest
from unittest.mock import MagicMock, patch, AsyncMock

from slidetextbridge.plugins import debug_helper

from test.filter_helper import *

class TestStdoutEmitter(unittest.IsolatedAsyncioTestCase):

    def test_type_name(self):
        self.assertEqual(debug_helper.StdoutEmitter.type_name(), 'stdout')

    async def test_stdout(self):

        with patch('sys.stdout.write') as mock_write:
            res = await run_filter(
                    debug_helper.StdoutEmitter,
                    {},
                    make_slide(['A', 'B'])
            )
            text = ''.join(x[0][0] for x in mock_write.call_args_list)
        self.assertEqual(text, 'A\nB\n\n')

        with patch('sys.stdout.write') as mock_write:
            res = await run_filter(
                    debug_helper.StdoutEmitter,
                    {'page_delimiter': 'C'},
                    make_slide(['A', 'B'])
            )
            text = ''.join(x[0][0] for x in mock_write.call_args_list)
        self.assertEqual(text, 'A\nBC')

    async def test_stdout_json(self):
        slide = make_slide(['A', 'B'])
        with patch('sys.stdout.write') as mock_write:
            res = await run_filter(
                    debug_helper.StdoutEmitter,
                    {'json': True},
                    slide
            )
            text = ''.join(x[0][0] for x in mock_write.call_args_list)
        self.assertEqual(json.loads(text), slide.to_dict())



import sys
import json
import unittest
from unittest.mock import MagicMock, patch, AsyncMock
import aiohttp

from slidetextbridge.plugins import openlp # pylint: disable=C0413

def _mock_poll(results):
    mockWSConnect = AsyncMock()
    mockWSConnect.return_value = AsyncMock()
    mockWSConnect.return_value.recv = AsyncMock()
    mockWSConnect.return_value.recv.return_value = json.dumps(
            {'results': results}
    )
    return patch('websockets.connect', mockWSConnect)

def _make_mock_request(data):
    mockResponse = AsyncMock()
    mockResponse.text.return_value = json.dumps(data)

    request_ctx = MagicMock()
    request_ctx.__aenter__ = AsyncMock(return_value=mockResponse)
    request_ctx.__aexit__ = AsyncMock(return_value=None)

    return MagicMock(return_value=request_ctx)


class TestOpenLPCapture(unittest.IsolatedAsyncioTestCase):
    def test_type_name(self):
        self.assertEqual(openlp.OpenLPCapture.type_name(), 'openlp')

    async def test_loop(self):
        ctx = MagicMock()
        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        obj._loop_once = AsyncMock()
        obj._loop_once.side_effect = (None, )
        obj._conn_ws = True
        obj.logger = MagicMock()
        with patch('asyncio.sleep', side_effect=(None, )) as mock_sleep:
            with self.assertRaises(StopAsyncIteration):
                await obj._loop()

        self.assertEqual(obj._loop_once.await_count, 3)
        self.assertEqual(mock_sleep.await_count, 2)

    async def test_text(self):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        poll_results = {
                'counter': 4, 'service': 0, 'slide': 0, 'item': 'd61e5b58-530d-11f0-9697-6045cba385ca', 'twelve': True, 'blank': False, 'theme': False, 'display': False, 'version': 3, 'isSecure': False, 'chordNotation': 'english'
        }

        data = {
            "title": "Song Title", "name": "songs", "type": 1, "theme": None, "footer": ["Song Title", "Written by: The Author"], "audit": ["Song Title", ["The Author"], "", ""], "notes": "", "data": {"title": "song title@alternate title", "alternate_title": "alternate title", "authors": "The Author", "ccli_number": "", "copyright": ""}, "fromPlugin": True, "capabilities": [2, 1, 5, 8, 9, 13, 22], "backgroundAudio": [], "isThemeOverwritten": False,
            "slides": [{"tag": "V1", "title": "Song Title", "selected": True, "text": "test text", "html": "lyrics 1", "chords": "lyrics 1", "footer": "Song Title<br/>    Written by:&nbsp;The Author<br/>"}, {"tag": "V1", "title": "Song Title", "selected": False, "text": "next text", "html": "lyrics 2", "chords": "lyrics 2", "footer": "Song Title<br/>    Written by:&nbsp;The Author<br/>"}], "id": "d61e5b58-530d-11f0-9697-6045cba385ca"
        }

        mockTCPConnector = MagicMock()
        mockRequest = _make_mock_request(data)

        obj.emit = AsyncMock()

        with (
            _mock_poll(poll_results),
            patch.multiple(
                'aiohttp',
                TCPConnector = mockTCPConnector,
                request = mockRequest
        )):
            await obj._loop_once()

        obj.emit.assert_called_once()
        slide_dict = obj.emit.call_args[0][0].to_dict()
        self.assertEqual(slide_dict['shapes'][0]['text'], 'test text')

    async def test_blank(self):
        await self._test_blank_types({'blank': True})
    async def test_theme(self):
        await self._test_blank_types({'blank': False, 'theme': True})
    async def test_display(self):
        await self._test_blank_types({'blank': False, 'theme': False,'display': True})

    async def _test_blank_types(self, results):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        mockTCPConnector = MagicMock()
        mockRequest = MagicMock()

        obj.emit = AsyncMock()

        with (
            _mock_poll(results),
            patch.multiple(
                'aiohttp',
                TCPConnector = mockTCPConnector,
                request = mockRequest
        )):
            await obj._loop_once()

        mockTCPConnector.assert_not_called()
        mockRequest.assert_not_called()
        obj.emit.assert_called_once()
        slide_dict = obj.emit.call_args[0][0].to_dict()
        self.assertEqual(slide_dict['shapes'], [])

    async def test_image(self):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        poll_results = {
                'counter': 4, 'service': 0, 'slide': 0, 'item': 'a121176e-5318-11f0-9697-6045cba385ca', 'twelve': True, 'blank': False, 'theme': False, 'display': False, 'version': 3, 'isSecure': False, 'chordNotation': 'english'
        }

        data = {
                'title': 'IMG_3685.heif',
                'name': 'images',
                'type': 2,
                'footer': [], 'audit': '',
                'notes': '',
                'data': {},
                'fromPlugin': True,
                'capabilities': [3, 1, 5, 6, 17, 21, 26],
                'slides': [{'tag': 1, 'title': 'an_image_file.heif', 'selected': True, 'img': 'data:image/heif;base64,'}],
                'id': 'a121176e-5318-11f0-9697-6045cba385ca',
        }

        mockTCPConnector = MagicMock()
        mockRequest = _make_mock_request(data)

        obj.emit = AsyncMock()

        with (
            _mock_poll(poll_results),
            patch.multiple(
                'aiohttp',
                TCPConnector = mockTCPConnector,
                request = mockRequest
        )):
            await obj._loop_once()

        obj.emit.assert_called_once()
        slide_texts = obj.emit.call_args[0][0].to_texts()
        self.assertEqual(slide_texts, [])

    async def test_poll_failure(self):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        obj.emit = AsyncMock()

        with _mock_poll(results=None) as mock_ws_connect:
            mock_ws_connect.side_effect = (OSError('test'), )
            with self.assertRaises(Exception):
                await obj._loop_once()

        obj.emit.assert_not_called()

    async def test_poll_response_failure(self):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        poll_results = {
                'counter': 4, 'service': 0, 'slide': 0, 'item': 'a121176e-5318-11f0-9697-6045cba385ca', 'twelve': True, 'blank': False, 'theme': False, 'display': False, 'version': 3, 'isSecure': False, 'chordNotation': 'english'
        }

        mockTCPConnector = MagicMock()
        mockRequest = AsyncMock()

        obj.emit = AsyncMock()
        obj.logger = MagicMock()

        with _mock_poll(results=None) as mock_ws_connect:
            mock_ws_connect.return_value.recv.side_effect = (OSError('test'), )
            await obj._loop_once()

        self.assertIn('Failed to poll.', obj.logger.warning.call_args[0][0])

        obj.emit.assert_called_once()
        slide_texts = obj.emit.call_args[0][0].to_texts()
        self.assertEqual(slide_texts, [])

    async def test_request_failure(self):
        ctx = MagicMock()

        cfg = openlp.OpenLPCapture.config({})
        obj = openlp.OpenLPCapture(ctx=ctx, cfg=cfg)

        poll_results = {
                'counter': 4, 'service': 0, 'slide': 0, 'item': 'a121176e-5318-11f0-9697-6045cba385ca', 'twelve': True, 'blank': False, 'theme': False, 'display': False, 'version': 3, 'isSecure': False, 'chordNotation': 'english'
        }

        mockTCPConnector = MagicMock()

        obj.emit = AsyncMock()
        obj.logger = MagicMock()

        with (
            _mock_poll(poll_results),
            patch('aiohttp.TCPConnector', mockTCPConnector),
            patch('aiohttp.request', side_effect=aiohttp.client_exceptions.ClientOSError('test-exception')),
        ):
            await obj._loop_once()

        self.assertIn('Failed to get live-items.', obj.logger.warning.call_args[0][0])

        obj.emit.assert_called_once()
        slide_texts = obj.emit.call_args[0][0].to_texts()
        self.assertEqual(slide_texts, [])


if __name__ == "__main__":
    unittest.main()
